@@ -36,6 +36,10 @@
 #include <QGuiApplication>
 #endif
 
+#if HAVE_XCB
+#include <xcb/xcb.h>
+#endif
+
 #include <QDBusConnectionInterface>
 #include <QDBusMessage>
 #include <QFile>
@@ -65,17 +69,11 @@
 // Dispose slaves after being idle for SLAVE_MAX_IDLE seconds
 #define SLAVE_MAX_IDLE  30
 
-// #define KLAUNCHER_VERBOSE_OUTPUT
-
 #include <QLoggingCategory>
 Q_DECLARE_LOGGING_CATEGORY(KLAUNCHER)
 
-#if QT_VERSION >= QT_VERSION_CHECK(5, 4, 0)
 // logging category for this framework, default: log stuff >= warning
 Q_LOGGING_CATEGORY(KLAUNCHER, "kf5.kinit.klauncher", QtWarningMsg)
-#else
-Q_LOGGING_CATEGORY(KLAUNCHER, "kf5.kinit.klauncher")
-#endif
 
 static const char *const s_DBusStartupTypeToString[] =
 { "DBusNone", "DBusUnique", "DBusMulti", "ERROR" };
@@ -102,7 +100,7 @@ ssize_t kde_safe_write(int fd, const void *buf, size_t count)
 
 #ifndef USE_KPROCESS_FOR_KIOSLAVES
 KLauncher::KLauncher(int _kdeinitSocket)
-    : QObject(0),
+    : QObject(nullptr),
       kdeinitSocket(_kdeinitSocket)
 #else
 KLauncher::KLauncher()
@@ -110,18 +108,15 @@ KLauncher::KLauncher()
 #endif
 {
 #if HAVE_X11
-    mCached_dpy = NULL;
     mIsX11 = QGuiApplication::platformName() == QStringLiteral("xcb");
 #endif
-    Q_ASSERT(g_klauncher_self == NULL);
+    Q_ASSERT(g_klauncher_self == nullptr);
     g_klauncher_self = this;
 
-    mAutoTimer.setSingleShot(true);
     new KLauncherAdaptor(this);
     mSlaveLauncherAdaptor = new KSlaveLauncherAdaptor(this);
     QDBusConnection::sessionBus().registerObject(QStringLiteral("/KLauncher"), this); // same as ktoolinvocation.cpp
 
-    connect(&mAutoTimer, SIGNAL(timeout()), this, SLOT(slotAutoStart()));
     connect(QDBusConnection::sessionBus().interface(),
             SIGNAL(serviceOwnerChanged(QString,QString,QString)),
             SLOT(slotNameOwnerChanged(QString,QString,QString)));
@@ -142,7 +137,7 @@ KLauncher::KLauncher()
             this, SLOT(slotKDEInitData(int)));
     kdeinitNotifier->setEnabled(true);
 #endif
-    lastRequest = 0;
+    lastRequest = nullptr;
     bProcessingQueue = false;
 
     mSlaveDebug = QString::fromLocal8Bit(qgetenv("KDE_SLAVE_DEBUG_WAIT"));
@@ -175,15 +170,15 @@ KLauncher::KLauncher()
 KLauncher::~KLauncher()
 {
     close();
-    g_klauncher_self = NULL;
+    g_klauncher_self = nullptr;
 }
 
 void KLauncher::close()
 {
-#if HAVE_X11
-    if (mCached_dpy != NULL) {
-        XCloseDisplay(mCached_dpy);
-        mCached_dpy = NULL;
+#if HAVE_XCB
+    if (mCached) {
+        xcb_disconnect(mCached.conn);
+        mCached = XCBConnection();
     }
 #endif
 }
@@ -236,7 +231,7 @@ read_socket(int sock, char *buffer, int len)
         timeval tm = { 30, 0 }; // 30 seconds timeout, so we're not stuck in case kdeinit dies on us
         FD_ZERO(&in);
         FD_SET(sock, &in);
-        select(sock + 1, &in, 0, 0, &tm);
+        select(sock + 1, &in, nullptr, nullptr, &tm);
         if (!FD_ISSET(sock, &in)) {
             qCDebug(KLAUNCHER) << "read_socket" << sock << "nothing to read, kdeinit5 must be dead";
             return -1;
@@ -305,7 +300,7 @@ void KLauncher::processRequestReturn(int status, const QByteArray &requestData)
             lastRequest->status = KLaunchRequest::Launching;
             break;
         }
-        lastRequest = 0;
+        lastRequest = nullptr;
         return;
     }
     if (lastRequest && (status == LAUNCHER_ERROR)) {
@@ -314,7 +309,7 @@ void KLauncher::processRequestReturn(int status, const QByteArray &requestData)
         if (!requestData.isEmpty()) {
             lastRequest->errorMsg = QString::fromUtf8((char *) requestData.data());
         }
-        lastRequest = 0;
+        lastRequest = nullptr;
         return;
     }
 
@@ -324,41 +319,27 @@ void KLauncher::processRequestReturn(int status, const QByteArray &requestData)
 void
 KLauncher::processDied(pid_t pid, long exitStatus)
 {
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
     qCDebug(KLAUNCHER) << pid << "exitStatus=" << exitStatus;
-#else
-    Q_UNUSED(exitStatus);
     // We should probably check the exitStatus for the uniqueapp case?
-#endif
     foreach (KLaunchRequest *request, requestList) {
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
         qCDebug(KLAUNCHER) << "  had pending request" << request->pid;
-#endif
         if (request->pid == pid) {
             if ((request->dbus_startup_type == KService::DBusUnique)
                     && QDBusConnection::sessionBus().interface()->isServiceRegistered(request->dbus_name)) {
                 request->status = KLaunchRequest::Running;
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
                 qCDebug(KLAUNCHER) << pid << "running as a unique app";
-#endif
             } else if(request->dbus_startup_type == KService::DBusNone && request->wait) {
                 request->status = KLaunchRequest::Running;
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
                 qCDebug(KLAUNCHER) << pid << "running as DBusNone with wait to true";
-#endif
             } else {
                 request->status = KLaunchRequest::Error;
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
                 qCDebug(KLAUNCHER) << pid << "died, requestDone. status=" << request->status;
-#endif
             }
             requestDone(request);
             return;
         }
     }
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
     qCDebug(KLAUNCHER) << "found no pending requests for PID" << pid;
-#endif
 }
 
 static bool matchesPendingRequest(const QString &appId, const QString &pendingAppId)
@@ -377,6 +358,11 @@ static bool matchesPendingRequest(const QString &appId, const QString &pendingAp
         return appName == pendingName;
     }
 
+    // Match sandboxed apps (e.g. flatpak), see https://phabricator.kde.org/D5775
+    if (newAppId.endsWith(QLatin1String(".kdbus"))) {
+        return newAppId.leftRef(newAppId.length() - 6) == pendingAppId;
+    }
+
     return newAppId == pendingAppId;
 }
 
@@ -389,46 +375,35 @@ KLauncher::slotNameOwnerChanged(const QString &appId, const QString &oldOwner,
         return;
     }
 
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
     qCDebug(KLAUNCHER) << "new app" << appId;
-#endif
     foreach (KLaunchRequest *request, requestList) {
         if (request->status != KLaunchRequest::Launching) {
             continue;
         }
 
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
         qCDebug(KLAUNCHER) << "had pending request" << request->name << s_DBusStartupTypeToString[request->dbus_startup_type] << "dbus_name" << request->dbus_name << request->tolerant_dbus_name;
-#endif
+
         // For unique services check the requested service name first
         if (request->dbus_startup_type == KService::DBusUnique) {
             if ((appId == request->dbus_name) || // just started
                     QDBusConnection::sessionBus().interface()->isServiceRegistered(request->dbus_name)) { // was already running
                 request->status = KLaunchRequest::Running;
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
                 qCDebug(KLAUNCHER) << "OK, unique app" << request->dbus_name << "is running";
-#endif
                 requestDone(request);
                 continue;
             } else {
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
                 qCDebug(KLAUNCHER) << "unique app" << request->dbus_name << "not running yet";
-#endif
             }
         }
 
         const QString rAppId = !request->tolerant_dbus_name.isEmpty() ? request->tolerant_dbus_name : request->dbus_name;
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
         qCDebug(KLAUNCHER) << "using" << rAppId << "for matching";
-#endif
         if (rAppId.isEmpty()) {
             continue;
         }
 
         if (matchesPendingRequest(appId, rAppId)) {
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
             qCDebug(KLAUNCHER) << "ok, request done";
-#endif
             request->dbus_name = appId;
             request->status = KLaunchRequest::Running;
             requestDone(request);
@@ -437,49 +412,6 @@ KLauncher::slotNameOwnerChanged(const QString &appId, const QString &oldOwner,
     }
 }
 
-void
-KLauncher::autoStart(int phase)
-{
-    if (mAutoStart.phase() >= phase) {
-        return;
-    }
-    mAutoStart.setPhase(phase);
-    if (phase == 0) {
-        mAutoStart.loadAutoStartList();
-    }
-    mAutoTimer.start(0);
-}
-
-void
-KLauncher::slotAutoStart()
-{
-    KService::Ptr s;
-    do {
-        QString service = mAutoStart.startService();
-        qCDebug(KLAUNCHER) << "Service: " << mAutoStart.phase() << service;
-        if (service.isEmpty()) {
-            // Done
-            if (!mAutoStart.phaseDone()) {
-                mAutoStart.setPhaseDone();
-                switch (mAutoStart.phase()) {
-                case 0:
-                    emit autoStart0Done();
-                    break;
-                case 1:
-                    emit autoStart1Done();
-                    break;
-                case 2:
-                    emit autoStart2Done();
-                    break;
-                }
-            }
-            return;
-        }
-        s = new KService(service);
-    } while (!start_service(s, QStringList(), QStringList(), "0", false, true, QDBusMessage()));
-    // Loop till we find a service that we can start.
-}
-
 void
 KLauncher::requestDone(KLaunchRequest *request)
 {
@@ -498,33 +430,18 @@ KLauncher::requestDone(KLaunchRequest *request)
         }
         requestResult.pid = 0;
 
-#if HAVE_X11
+#if HAVE_XCB
         if (!request->startup_dpy.isEmpty() && mIsX11) {
-            Display *dpy = NULL;
-            if ((mCached_dpy != NULL) &&
-                    (request->startup_dpy == XDisplayString(mCached_dpy))) {
-                dpy = mCached_dpy;
-            }
-            if (dpy == NULL) {
-                dpy = XOpenDisplay(request->startup_dpy.constData());
-            }
-            if (dpy) {
+            XCBConnection conn = getXCBConnection(request->startup_dpy);
+            if (conn) {
                 KStartupInfoId id;
                 id.initId(request->startup_id);
-                KStartupInfo::sendFinishX(dpy, id);
-                if (mCached_dpy != dpy && mCached_dpy != NULL) {
-                    XCloseDisplay(mCached_dpy);
-                }
-                mCached_dpy = dpy;
+                KStartupInfo::sendFinishXcb(conn.conn, conn.screen, id);
             }
         }
 #endif
     }
 
-    if (request->autoStart) {
-        mAutoTimer.start(0);
-    }
-
     if (request->transaction.type() != QDBusMessage::InvalidMessage) {
         if (requestResult.dbusName.isNull()) { // null strings can't be sent
             requestResult.dbusName.clear();
@@ -536,10 +453,8 @@ KLauncher::requestDone(KLaunchRequest *request)
                                            << requestResult.error
                                            << stream_pid));
     }
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
-    qCDebug(KLAUNCHER) << "removing done request" << request->name << "PID" << request->pid;
-#endif
 
+    qCDebug(KLAUNCHER) << "removing done request" << request->name << "PID" << request->pid;
     requestList.removeAll(request);
     delete request;
 }
@@ -564,14 +479,22 @@ KLauncher::requestStart(KLaunchRequest *request)
     connect(process, SIGNAL(finished(int,QProcess::ExitStatus)), this, SLOT(slotFinished(int,QProcess::ExitStatus)));
     request->process = process;
 
-// process.setEnvironment(envlist);
+    QProcessEnvironment environment = QProcessEnvironment::systemEnvironment();
+    foreach (const QString &env, request->envs) {
+        const int pos = env.indexOf(QLatin1Char('='));
+        const QString envVariable = env.left(pos);
+        const QString envValue = env.mid(pos + 1);
+        environment.insert(envVariable, envValue);
+    }
+    process->setProcessEnvironment(environment);
+
     QStringList args;
     foreach (const QString &arg, request->arg_list) {
         args << arg;
     }
 
     QString executable = request->name;
-#ifdef Q_OS_MAC
+#ifdef Q_OS_OSX
     const QString bundlepath = QStandardPaths::findExecutable(executable);
     if (!bundlepath.isEmpty()) {
         executable = bundlepath;
@@ -627,10 +550,8 @@ KLauncher::requestStart(KLaunchRequest *request)
 #endif
     request_header.arg_length = requestData.length();
 
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
     qCDebug(KLAUNCHER) << "Asking kdeinit to start" << request->name << request->arg_list
             << "cmd=" << commandToString(request_header.cmd);
-#endif
 
     kde_safe_write(kdeinitSocket, &request_header, sizeof(request_header));
     kde_safe_write(kdeinitSocket, requestData.data(), requestData.length());
@@ -639,14 +560,13 @@ KLauncher::requestStart(KLaunchRequest *request)
     lastRequest = request;
     do {
         slotKDEInitData(kdeinitSocket);
-    } while (lastRequest != 0);
+    } while (lastRequest != nullptr);
 #endif
 }
 
 void KLauncher::exec_blind(const QString &name, const QStringList &arg_list, const QStringList &envs, const QString &startup_id)
 {
     KLaunchRequest *request = new KLaunchRequest;
-    request->autoStart = false;
     request->name = name;
     request->arg_list =  arg_list;
     request->dbus_startup_type = KService::DBusNone;
@@ -686,10 +606,10 @@ KLauncher::start_service_by_desktop_path(const QString &serviceName, const QStri
     if (!service) {
         requestResult.result = ENOENT;
         requestResult.error = i18n("Could not find service '%1'.", serviceName);
-        cancel_service_startup_info(NULL, startup_id.toLocal8Bit(), envs);   // cancel it if any
+        cancel_service_startup_info(nullptr, startup_id.toLocal8Bit(), envs);   // cancel it if any
         return false;
     }
-    return start_service(service, urls, envs, startup_id.toLocal8Bit(), blind, false, msg);
+    return start_service(service, urls, envs, startup_id.toLocal8Bit(), blind, msg);
 }
 
 bool
@@ -700,16 +620,16 @@ KLauncher::start_service_by_desktop_name(const QString &serviceName, const QStri
     if (!service) {
         requestResult.result = ENOENT;
         requestResult.error = i18n("Could not find service '%1'.", serviceName);
-        cancel_service_startup_info(NULL, startup_id.toLocal8Bit(), envs);   // cancel it if any
+        cancel_service_startup_info(nullptr, startup_id.toLocal8Bit(), envs);   // cancel it if any
         return false;
     }
-    return start_service(service, urls, envs, startup_id.toLocal8Bit(), blind, false, msg);
+    return start_service(service, urls, envs, startup_id.toLocal8Bit(), blind, msg);
 }
 
 bool
 KLauncher::start_service(KService::Ptr service, const QStringList &_urls,
                          const QStringList &envs, const QByteArray &startup_id,
-                         bool blind, bool autoStart, const QDBusMessage &msg)
+                         bool blind, const QDBusMessage &msg)
 {
     QStringList urls = _urls;
     bool runPermitted = KDesktopFile::isAuthorizedDesktopFile(service->entryPath());
@@ -721,11 +641,35 @@ KLauncher::start_service(KService::Ptr service, const QStringList &_urls,
         } else {
             requestResult.error = i18n("Service '%1' is malformatted.", service->entryPath());
         }
-        cancel_service_startup_info(NULL, startup_id, envs);   // cancel it if any
+        cancel_service_startup_info(nullptr, startup_id, envs);   // cancel it if any
         return false;
     }
     KLaunchRequest *request = new KLaunchRequest;
-    request->autoStart = autoStart;
+
+    enum DiscreteGpuCheck { NotChecked, Present, Absent };
+    static DiscreteGpuCheck s_gpuCheck = NotChecked;
+
+    if (service->runOnDiscreteGpu() && s_gpuCheck == NotChecked) {
+        // Check whether we have a discrete gpu
+        bool hasDiscreteGpu = false;
+        QDBusInterface iface(QLatin1String("org.kde.Solid.PowerManagement"),
+                             QLatin1String("/org/kde/Solid/PowerManagement"),
+                             QLatin1String("org.kde.Solid.PowerManagement"),
+                             QDBusConnection::sessionBus());
+        if (iface.isValid()) {
+            QDBusReply<bool> reply = iface.call(QLatin1String("hasDualGpu"));
+            if (reply.isValid()) {
+                hasDiscreteGpu = reply.value();
+            }
+        }
+
+        s_gpuCheck = hasDiscreteGpu ? Present : Absent;
+    }
+
+    QStringList _envs = envs;
+    if (service->runOnDiscreteGpu() && s_gpuCheck == Present) {
+        _envs << QLatin1String("DRI_PRIME=1");
+    }
 
     if ((urls.count() > 1) && !service->allowMultipleFiles()) {
         // We need to launch the application N times. That sucks.
@@ -743,17 +687,13 @@ KLauncher::start_service(KService::Ptr service, const QStringList &_urls,
             if (!startup_id2.isEmpty() && startup_id2 != "0") {
                 startup_id2 = "0";    // can't use the same startup_id several times // krazy:exclude=doublequote_chars
             }
-            start_service(service, singleUrl, envs, startup_id2, true, false, msg);
+            start_service(service, singleUrl, _envs, startup_id2, true, msg);
         }
-        QString firstURL = *(urls.begin());
+        const QString firstURL = urls.at(0);
         urls.clear();
         urls.append(firstURL);
     }
-    // Qt5 TODO: use QUrl::fromStringList
-    QList<QUrl> qurls;
-    Q_FOREACH (const QString &u, urls) {
-        qurls.append(QUrl(u));
-    }
+    const QList<QUrl> qurls = QUrl::fromStringList(urls);
 
     createArgs(request, service, qurls);
 
@@ -762,7 +702,7 @@ KLauncher::start_service(KService::Ptr service, const QStringList &_urls,
         requestResult.result = ENOEXEC;
         requestResult.error = i18n("Service '%1' is malformatted.", service->entryPath());
         delete request;
-        cancel_service_startup_info(NULL, startup_id, envs);
+        cancel_service_startup_info(nullptr, startup_id, _envs);
         return false;
     }
 
@@ -792,18 +732,16 @@ KLauncher::start_service(KService::Ptr service, const QStringList &_urls,
         }
     }
 
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
     qCDebug(KLAUNCHER) << "name=" << request->name << "dbus_name=" << request->dbus_name
             << "startup type=" << s_DBusStartupTypeToString[request->dbus_startup_type];
-#endif
 
     request->pid = 0;
     request->wait = false;
-    request->envs = envs;
-    send_service_startup_info(request, service, startup_id, envs);
+    request->envs = _envs;
+    send_service_startup_info(request, service, startup_id, _envs);
 
     // Request will be handled later.
-    if (!blind && !autoStart) {
+    if (!blind) {
         msg.setDelayedReply(true);
         request->transaction = msg;
     }
@@ -815,7 +753,7 @@ void
 KLauncher::send_service_startup_info(KLaunchRequest *request, KService::Ptr service, const QByteArray &startup_id,
                                      const QStringList &envs)
 {
-#if HAVE_X11
+#if HAVE_XCB
     if (!mIsX11) {
         return;
     }
@@ -836,20 +774,15 @@ KLauncher::send_service_startup_info(KLaunchRequest *request, KService::Ptr serv
             dpy_str = env.mid(8).toLocal8Bit();
         }
     }
-    Display *dpy = NULL;
-    if (!dpy_str.isEmpty() && mCached_dpy != NULL && dpy_str != XDisplayString(mCached_dpy)) {
-        dpy = mCached_dpy;
-    }
-    if (dpy == NULL) {
-        dpy = XOpenDisplay(dpy_str.constData());
-    }
+
+    XCBConnection conn = getXCBConnection(dpy_str);
     request->startup_id = id.id();
-    if (dpy == NULL) {
+    if (!conn) {
         cancel_service_startup_info(request, startup_id, envs);
         return;
     }
 
-    request->startup_dpy = dpy_str;
+    request->startup_dpy = conn.displayName;
 
     KStartupInfoData data;
     data.setName(service->name());
@@ -863,11 +796,7 @@ KLauncher::send_service_startup_info(KLaunchRequest *request, KService::Ptr serv
     }
     data.setApplicationId(service->entryPath());
     // the rest will be sent by kdeinit
-    KStartupInfo::sendStartupX(dpy, id, data);
-    if (mCached_dpy != dpy && mCached_dpy != NULL) {
-        XCloseDisplay(mCached_dpy);
-    }
-    mCached_dpy = dpy;
+    KStartupInfo::sendStartupXcb(conn.conn, conn.screen, id, data);
     return;
 #else
     return;
@@ -878,8 +807,8 @@ void
 KLauncher::cancel_service_startup_info(KLaunchRequest *request, const QByteArray &startup_id,
                                        const QStringList &envs)
 {
-#if HAVE_X11
-    if (request != NULL) {
+#if HAVE_XCB
+    if (request != nullptr) {
         request->startup_id = "0";    // krazy:exclude=doublequote_chars
     }
     if (!startup_id.isEmpty() && startup_id != "0" && mIsX11) {
@@ -889,24 +818,13 @@ KLauncher::cancel_service_startup_info(KLaunchRequest *request, const QByteArray
                 dpy_str = env.mid(8);
             }
         }
-        Display *dpy = NULL;
-        if (!dpy_str.isEmpty() && mCached_dpy != NULL
-                && dpy_str != QLatin1String(XDisplayString(mCached_dpy))) {
-            dpy = mCached_dpy;
-        }
-        if (dpy == NULL) {
-            dpy = XOpenDisplay(dpy_str.toLatin1().constData());
-        }
-        if (dpy == NULL) {
+        XCBConnection conn = getXCBConnection(dpy_str.toLocal8Bit());
+        if (!conn) {
             return;
         }
         KStartupInfoId id;
         id.initId(startup_id);
-        KStartupInfo::sendFinishX(dpy, id);
-        if (mCached_dpy != dpy && mCached_dpy != NULL) {
-            XCloseDisplay(mCached_dpy);
-        }
-        mCached_dpy = dpy;
+        KStartupInfo::sendFinishXcb(conn.conn, conn.screen, id);
     }
 #endif
 }
@@ -917,7 +835,6 @@ KLauncher::kdeinit_exec(const QString &app, const QStringList &args,
                         const QString &startup_id, bool wait, const QDBusMessage &msg)
 {
     KLaunchRequest *request = new KLaunchRequest;
-    request->autoStart = false;
     request->arg_list = args;
     request->name = app;
     request->dbus_startup_type = KService::DBusNone;
@@ -967,9 +884,7 @@ KLauncher::slotDequeue()
         requestStart(request);
         if (request->status != KLaunchRequest::Launching) {
             // Request handled.
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
             qCDebug(KLAUNCHER) << "Request handled already";
-#endif
             requestDone(request);
             continue;
         }
@@ -1004,7 +919,7 @@ pid_t
 KLauncher::requestHoldSlave(const QString &urlStr, const QString &app_socket)
 {
     const QUrl url(urlStr);
-    IdleSlave *slave = 0;
+    IdleSlave *slave = nullptr;
     foreach (IdleSlave *p, mSlaveList) {
         if (p->onHold(url)) {
             slave = p;
@@ -1025,7 +940,7 @@ KLauncher::requestSlave(const QString &protocol,
                         const QString &app_socket,
                         QString &error)
 {
-    IdleSlave *slave = 0;
+    IdleSlave *slave = nullptr;
     foreach (IdleSlave *p, mSlaveList) {
         if (p->match(protocol, host, true)) {
             slave = p;
@@ -1072,7 +987,11 @@ KLauncher::requestSlave(const QString &protocol,
     arg_list << protocol;
     arg_list << mConnectionServer.address().toString();
     arg_list << app_socket;
+#ifdef Q_OS_WIN
+    QString name = QLatin1String("kioslave");
+#else
     QString name = QFile::decodeName(CMAKE_INSTALL_FULL_LIBEXECDIR_KF5 "/kioslave");
+#endif
 #else
     QString arg1 = protocol;
     QString arg2 = mConnectionServer.address().toString();
@@ -1098,19 +1017,22 @@ KLauncher::requestSlave(const QString &protocol,
     if (mSlaveValgrind == protocol) {
         arg_list.prepend(name);
 #ifndef USE_KPROCESS_FOR_KIOSLAVES // otherwise we've already done this
+#ifdef Q_OS_WIN
+        arg_list.prepend(QLatin1String("kioslave"));
+#else
         arg_list.prepend(QFile::decodeName(CMAKE_INSTALL_FULL_LIBEXECDIR_KF5 "/kioslave"));
+#endif
 #endif
         name = QStringLiteral("valgrind");
 
         if (!mSlaveValgrindSkin.isEmpty()) {
             arg_list.prepend(QLatin1String("--tool=") + mSlaveValgrindSkin);
         } else {
-            arg_list.prepend(QStringLiteral("--tool=memcheck"));
+            arg_list.prepend(QLatin1String("--tool=memcheck"));
         }
     }
 #endif
     KLaunchRequest *request = new KLaunchRequest;
-    request->autoStart = false;
     request->name = name;
     request->arg_list =  arg_list;
     request->dbus_startup_type = KService::DBusNone;
@@ -1128,7 +1050,7 @@ KLauncher::requestSlave(const QString &protocol,
     // We don't care about this request any longer....
     requestDone(request);
     if (!pid) {
-        error = i18n("Error loading '%1'.\n", name);
+        error = i18n("Error loading '%1'.", name);
     }
     return pid;
 }
@@ -1186,6 +1108,11 @@ KLauncher::slotSlaveStatus(IdleSlave *slave)
             delete waitRequest;
         }
     }
+
+    if (slave->hasTempAuthorization()) {
+        mSlaveList.removeAll(slave);
+        slave->deleteLater();
+    }
 }
 
 void
@@ -1240,9 +1167,7 @@ KLauncher::slotFinished(int exitCode, QProcess::ExitStatus exitStatus)
 
     foreach (KLaunchRequest *request, requestList) {
         if (request->process == p) {
-#ifdef KLAUNCHER_VERBOSE_OUTPUT
             qCDebug(KLAUNCHER) << "found KProcess, request done";
-#endif
             if (exitCode == 0  && exitStatus == QProcess::NormalExit) {
                 request->status = KLaunchRequest::Done;
             } else {
@@ -1270,3 +1195,36 @@ void KLauncher::terminate_kdeinit()
 #endif
 }
 
+#if HAVE_XCB
+KLauncher::XCBConnection KLauncher::getXCBConnection(const QByteArray &_displayName)
+{
+    const auto displayName = !_displayName.isEmpty() ? _displayName : qgetenv("DISPLAY");
+
+    // If cached connection is same as request
+    if (mCached && displayName == mCached.displayName) {
+        // Check error, if so close it, otherwise it's still valid, reuse the cached one
+        if (xcb_connection_has_error(mCached.conn)) {
+            close();
+        } else {
+            return mCached;
+        }
+    }
+
+    // At this point, the cached connection can't be reused, make a new connection
+    XCBConnection conn;
+    conn.conn = xcb_connect(displayName.constData(), &conn.screen);
+    if (conn) {
+        // check error first, if so return an empty one
+        if (xcb_connection_has_error(conn.conn)) {
+            xcb_disconnect(conn.conn);
+            return XCBConnection();
+        }
+
+        // if it's valid, replace mCached with new connection
+        conn.displayName = displayName;
+        close();
+        mCached = conn;
+    }
+    return conn;
+}
+#endif
